"""Old backends no longer in active development.

These might or might not work.
"""

from t2kdm.backends import *

class LCGBackend(GridBackend):
    """Grid backend using the LCG command line tools `lfc-*` and `lcg-*`."""

    def __init__(self, **kwargs):
        GridBackend.__init__(self, catalogue_prefix='lfn:/grid', **kwargs)

        #self._proxy_init_cmd = sh.Command('voms-proxy-init')
        self._ls_cmd = sh.Command('lfc-ls')
        self._replicas_cmd = sh.Command('lcg-lr')
        self._replica_state_cmd = sh.Command('lcg-ls')
        self._replica_checksum_cmd = sh.Command('lcg-get-checksum')
        self._bringonline_cmd = sh.Command('lcg-bringonline')
        self._replicate_cmd = sh.Command('lcg-rep')
        self._cp_cmd = sh.Command('lcg-cp')
        self._cr_cmd = sh.Command('lcg-cr')
        self._del_cmd = sh.Command('lcg-del')

    def _ls(self, lurl, **kwargs):
        # Translate keyword arguments
        d = kwargs.pop('directory', False)
        args = []
        if -d:
            args.append('-d')
        args.append('-l')
        args.append(lurl[4:])
        try:
            output = self._ls_cmd(*args, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or Directory.")
            else:
                raise
        for line in output:
            fields = line.split()
            mode, links, uid, gid, size = fields[:5]
            name = fields[-1]
            modified = ' '.join(fields[5:-1])
            yield DirEntry(name, mode=mode, links=int(links), gid=gid, uid=uid, size=int(size), modified=modified)

    def _replicas(self, lurl, **kwargs):
        ret = []
        try:
            output = self._replicas_cmd(lurl, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or Directory.")
        for line in output:
            line = line.strip()
            if len(line) > 0:
                ret.append(line.strip())
        return ret

    def _state(self, surl, **kwargs):
        it = kwargs.pop('_iter', None)
        try:
            listing = self._replica_state_cmd('-l', surl, **kwargs)
        except sh.ErrorReturnCode:
            listing = '- - - - - ?'
        except sh.SignalException_SIGSEGV:
            listing = '- - - - - ?'
        return listing.split()[5]

    def _checksum(self, surl, **kwargs):
        it = kwargs.pop('_iter', None)
        try:
            listing = self._replica_checksum_cmd(surl, **kwargs)
        except sh.ErrorReturnCode:
            listing = '? -'
        except sh.SignalException_SIGSEGV:
            listing = '? -'
        try:
            checksum = listing.split()[0]
        except IndexError:
            # Something weird happened
            checksum = '?'
        return checksum

    def _bringonline(self, surl, timeout, **kwargs):
        kwargs['_err_to_out'] = True # Verbose output is on stderr
        kwargs.pop('_err', None) # Cannot specify _err and _err_ro_out at same time

        # Get original command output
        return self._bringonline_cmd('-v', '--bdii-timeout', timeout, '--srm-timeout', timeout, '--sendreceive-timeout', timeout, '--connect-timeout', timeout, surl, **kwargs)

    def _replicate(self, source_surl, destination_surl, lurl, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._replicate_cmd('-v', '--sendreceive-timeout', 14400, '--checksum', '-d', destination_surl, source_surl, _out=out, _err_to_out=True, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        return True

    def _get(self, surl, localpath, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._cp_cmd('-v', '--sendreceive-timeout', 14400, '--checksum', surl, localpath, _out=out, _err_to_out=True, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        return os.path.isfile(localpath)


    def _put(self, localpath, surl, lurl, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._cr_cmd('-v', '--sendreceive-timeout', 14400, '--checksum', '-d', surl, '-l', lurl, localpath, _out=out, _err_to_out=True, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        return True

    def _remove(self, surl, lurl, last=False, verbose=True, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            if deregister:
                raise BackendException("Operation not supported by LCG backend.")
            else:
                self._del_cmd('-v', surl, _out=out, _err_to_out=True, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        return True

class GFALBackend(GridBackend):
    """Grid backend using the GFAL command line tools `gfal-*`."""

    def __init__(self, **kwargs):
        GridBackend.__init__(self, catalogue_prefix='lfn:/grid', **kwargs)

        #self._proxy_init_cmd = sh.Command('voms-proxy-init')
        self._ls_cmd = sh.Command('gfal-ls').bake(color='never')
        self._replicas_cmd = sh.Command('gfal-xattr')
        self._replica_checksum_cmd = sh.Command('gfal-sum')
        self._bringonline_cmd = sh.Command('gfal-legacy-bringonline')
        self._cp_cmd = sh.Command('gfal-copy')
        self._register_cmd = sh.Command('gfal-legacy-register')
        self._deregister_cmd = sh.Command('gfal-legacy-unregister')
        self._del_cmd = sh.Command('gfal-rm')

    def _ls(self, lurl, **kwargs):
        # Translate keyword arguments
        d = kwargs.pop('directory', False)
        args = []
        if -d:
            args.append('-d')
        args.append('-l')
        args.append(lurl)
        try:
            output = self._ls_cmd(*args, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or Directory.")
            else:
                raise BackendException(e.stderr)
        for line in output:
            fields = line.split()
            mode, links, gid, uid, size = fields[:5]
            name = fields[-1]
            modified = ' '.join(fields[5:-1])
            yield DirEntry(name, mode=mode, links=int(links), gid=gid, uid=uid, size=int(size), modified=modified)

    def _ls_se(self, surl, **kwargs):
        return self._ls(surl, **kwargs)

    def _replicas(self, lurl, **kwargs):
        ret = []
        try:
            output = self._replicas_cmd(lurl, 'user.replicas', **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or Directory.")
            else:
                raise BackendException(e.stderr)
        for line in output:
            line = line.strip()
            if len(line) > 0:
                ret.append(line.strip())
        return ret

    def _exists(self, surl, **kwargs):
        try:
            state = self._replicas_cmd(surl, 'user.status', **kwargs).strip()
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                return False
            else:
                raise BackendException(e.stderr)
        else:
            return True

    def _deregister(self, surl, lurl, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._deregister_cmd(lurl, surl, _out=out, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        else:
            return True

    def _state(self, surl, **kwargs):
        try:
            state = self._replicas_cmd(surl, 'user.status', **kwargs).strip()
        except sh.ErrorReturnCode:
            state = '?'
        except sh.SignalException_SIGSEGV:
            state = '?'
        return state

    def _checksum(self, surl, **kwargs):
        try:
            checksum = self._replica_checksum_cmd(surl, 'ADLER32', **kwargs).split()[1]
        except sh.ErrorReturnCode:
            checksum = '?'
        except sh.SignalException_SIGSEGV:
            checksum = '?'
        except IndexError:
            checksum = '?'
        return checksum

    def _bringonline(self, surl, timeout, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        # gfal does not notice when files come online, it seems
        # split task into many requests with short timeouts
        if verbose:
            out = sys.stdout
        else:
            out = None
        time_left = timeout
        while(True):
            if time_left > 10:
                timeout = 10
            else:
                timeout = time_left
            time_left -= 10
            try:
                self._bringonline_cmd('-t', timeout, surl, _out=out, **kwargs)
            except sh.ErrorReturnCode:
                # Not online yet.
                if time_left > 0:
                    continue
                else:
                    return False
            else:
                # File is online.
                return True

    def _replicate(self, source_surl, destination_surl, lurl, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._cp_cmd('-p', '-T', '1800', '--checksum', 'ADLER32', source_surl, destination_surl, _out=out, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            elif 'File exists' in e.stderr:
                if verbose:
                    print_("Replica already exists. Checking checksum...")
                if self.checksum(destination_surl) == self.checksum(source_surl):
                    if verbose:
                        print_("Checksums match. Registering replica.")
                else:
                    raise BackendException("File with different checksum already present.")
            else:
                raise BackendException(e.stderr)

        try:
            self._register_cmd(lurl, destination_surl, _out=out, **kwargs)
        except sh.ErrorReturnCode as e:
            raise BackendException(e.stderr)

        return True

    def _get(self, surl, localpath, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._cp_cmd('-f', '--checksum', 'ADLER32', surl, localpath, _out=out, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        return os.path.isfile(localpath)

    def _put(self, localpath, surl, lurl, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._cp_cmd('-p', '--checksum', 'ADLER32', localpath, surl, lurl, _out=out, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        return True

    def _remove(self, surl, lurl, last=False, verbose=False, **kwargs):
        if verbose:
            out = sys.stdout
        else:
            out = None
        try:
            self._del_cmd(surl, _out=out, **kwargs)
            self._deregister_cmd(lurl, surl, _out=out, **kwargs)
            if last:
                # Delete lfn
                self._del_cmd(lurl, _out=out, **kwargs)
        except sh.ErrorReturnCode as e:
            if 'No such file' in e.stderr:
                raise DoesNotExistException("No such file or directory.")
            else:
                raise BackendException(e.stderr)
        return True




"""Module handling the CLI and stand-alone script commands."""

from six import print_
import sh
import shlex
import argparse
import t2kdm
import t2kdm.interactive
import sys
from os import path
import posixpath

all_commands = []

class Command(object):
    """Commands to be used as stand alone scripts and within the CLI.

    Command line arguments are translated to the underlying function's
    positional and keyword arguments. Keyword arguments *must* define a short
    and long version.

    Example:

        ls = Command('ls', t2kdm.interactive.ls, "List contents of a remote logical path.")
        ls.add_argument('remotepath', type=str,
            help="the logical path, e.g. '/nd280'")
        ls.add_argument('-l', '--long', action='store_true',
            help="longer, more detailed output")

    The arguments containing `localpath` or `remotepath` are treated specially.
    These arguments are converted into an absolute path if the `localdir` or
    `remotedir` is specified repectively.
    """

    def __init__(self, name, function, description="", **kwargs):
        self.name = name
        self.function = function
        # Set prog to command name, iff we are running in the CLI
        if "t2kdm-cli" in sys.argv[0]:
            self.parser = argparse.ArgumentParser(prog=name, description=description, **kwargs)
        else:
            self.parser = argparse.ArgumentParser(description=description, **kwargs)
        self.positional_arguments = []
        self.keyword_arguments = []

    def add_argument(self, *args, **kwargs):
        """Add an argument to the parser and memorize how to pass parsed object to the original function.

        All positional arguments are identified by their first argument *not* starting with a '-'.
        They are passed to the original function as `*args`.

        All keyword arguments are identified by their first argument starting with a '-'.
        Their keys are taken from the second argument minus the leading '--'.
        They are passed to the original function as `**kwargs`.
        """

        if args[0][0] == '-':
            # Keyword argument
            self.keyword_arguments.append(args[1][2:])
        else:
            # Positionsl argument
            self.positional_arguments.append(args[0])

        # Add the arguments to the parser
        self.parser.add_argument(*args, **kwargs)

    def run_from_console(self, **kwargs):
        """Entry point for console scripts.

        Parses command line arguments from sys.argv.
        Prints output on screen.
        Returns error code (0 if successful).
        """

        try:
            ret = self.run(self.parser.parse_args(), **kwargs)
        except t2kdm.backends.BackendException as e:
            print_(e, file=sys.stderr)
            return 1
        except t2kdm.interactive.InteractiveException as e:
            print_(e, file=sys.stderr)
            return 1
        except IOError as e:
            if e.errno == 32:
                # A broken pipe, e.g. from using a command with `head`.
                # Throwing an exception for this looks untidy, so we just return an error code.
                return 1
            else:
                raise

        return ret

    def run_from_cli(self, argstring, _return=False, **kwargs):
        """Run command from command line interface.

        Parses single string into arguments.
        Prints output on screen.
        Returns `False`, unless `_return` is `True`.
        Then it returns the actual return value.
        """

        try:
            args = shlex.split(argstring)
        except ValueError as e: # Catch errors from bad bash syntax
            print_(e)
            return False

        ret = False
        try: # We do *not* want to exit after printing a help message or erroring, so we have to catch that.
            ret = self.run_from_arglist(args, **kwargs)
        except t2kdm.backends.BackendException as e:
            print_(e)
        except t2kdm.interactive.InteractiveException as e:
            print_(e)
        except Exception as e:
            print_(e)
        except SystemExit:
            pass

        if _return:
            return ret
        else:
            return False

    def run_from_arglist(self, arglist, **kwargs):
        """Run command with list of arguments."""
        args = self.parser.parse_args(arglist)
        return self.run(args, **kwargs)

    @staticmethod
    def _condition_argument(name, value, localdir=None, remotedir=None):
        """Apply some processing to the arguments when needed."""

        # Make local paths absolute
        if localdir is not None and 'localpath' in name and not path.isabs(value):
            value = path.normpath(path.join(localdir, value))

        # Make remote paths absolute
        if remotedir is not None and 'remotepath' in name and not posixpath.isabs(value):
            value = posixpath.normpath(posixpath.join(remotedir, value))

        # Special case: remote basename == @
        if 'remotepath' in name:
            # Get all entries in the directory and replace the @ with the (lexigraphically) last one
            dirname, basename = posixpath.split(value)
            if basename == '@':
                entries = [x.name for x in t2kdm.ls(dirname)]
                entries.sort()
                value = posixpath.join(dirname, entries[-1])

        return value

    def run(self, parsed_args, **kwargs):
        """Run the command.

        Takes an parsed arguments object from argparse as input.
        """

        localdir = kwargs.pop('localdir', None)
        remotedir = kwargs.pop('remotedir', None)

        pos_args = []
        for arg in self.positional_arguments:
            value = self._condition_argument(arg, getattr(parsed_args, arg), localdir=localdir, remotedir=remotedir)
            pos_args.append(value)

        key_args = {}
        for arg in self.keyword_arguments:
            value = self._condition_argument(arg, getattr(parsed_args, arg), localdir=localdir, remotedir=remotedir)
            key_args[arg] = value
        key_args.update(kwargs)

        return self(*pos_args, **key_args)

    def __call__(self, *args, **kwargs):
        """Call the underlying function directly."""
        return self.function(*args, **kwargs)

ls = Command('ls', t2kdm.interactive.ls, "List contents of a remote logical path.")
ls.add_argument('remotepath', type=str, nargs='?', default='',
    help="the remote logical path, e.g. '/nd280'")
ls.add_argument('-l', '--long', action='store_true',
    help="longer, more detailed output")
ls.add_argument('-d', '--directory', action='store_true',
    help="list directory entries instead of contents")
ls.add_argument('-s', '--se', type=str, default=None,
    help="list physical contents on this SE rather than the file catalogue")
all_commands.append(ls)

replicas = Command('replicas', t2kdm.interactive.replicas, "List replicas of a remote logical path.")
replicas.add_argument('remotepath', type=str, nargs='?', default='',
    help="the remote logical path, e.g. '/nd280/file.txt'")
replicas.add_argument('-c', '--checksum', action='store_true',
    help="display checksums of all replicas")
replicas.add_argument('-s', '--state', action='store_true',
    help="display the state of all replicas, e.g. 'ONLINE'")
replicas.add_argument('-n', '--name', action='store_true',
    help="display the name of the storage element")
all_commands.append(replicas)

check = Command('check', t2kdm.interactive.check, "Check the replicas of a given file/directory.")
check.add_argument('remotepath', type=str,
    help="the remote logical path, e.g. '/nd280'")
check.add_argument('-r', '--recursive', nargs='?', metavar="REGEX", default=False, const=True,
    help="recursively check all files and subdirectories [that match REGEX] of a directory")
check.add_argument('-R', '--recursivese', default=None,
    help="use listing of physical files on this SE for recursion")
check.add_argument('-q', '--quiet', action='store_true',
    help="do not print problematic files to screen")
check.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
check.add_argument('-l', '--list', metavar='FILENAME',
    help="save a list of failed files to FILENAME")
check.add_argument('-c', '--checksum', action='store_true',
    help="check whether the checksums of all replicas is identical, takes longer than just the se tests")
check.add_argument('-s', '--se', action='append', default=[],
    help="report replication status to the given storage element, can be used multiple times")
check.add_argument('-S', '--states', action='store_true',
    help="checl whether all replicas are in a resonable state ('ONLINE', 'NEARLINE', or 'ONLINE_AND_NEARLINE')")
all_commands.append(check)

replicate = Command('replicate', t2kdm.interactive.replicate, "Replicate file to a storage element.")
replicate.add_argument('remotepath', type=str,
    help="the remote logical path, e.g. '/nd280/file.txt'")
replicate.add_argument('destination', type=str,
    help="the destination storage element by name, e.g. 'UKI-SOUTHGRID-RALPP-disk', or by host, e.g. 't2ksrm.nd280.org'")
replicate.add_argument('-r', '--recursive', nargs='?', metavar="REGEX", default=False, const=True,
    help="recursively replicate all files and subdirectories [that match REGEX] of a directory")
replicate.add_argument('-R', '--recursivese', default=None,
    help="use listing of physical files on this SE for recursion")
replicate.add_argument('-l', '--list', metavar='FILENAME',
    help="save a list of failed files to FILENAME")
replicate.add_argument('-s', '--source', type=str, default=None,
    help="the source storage element by name, e.g. 'UKI-SOUTHGRID-RALPP-disk', or by host, e.g. 't2ksrm.nd280.org'. If no source is provided, the replica closest to the destination is chosen")
replicate.add_argument('-t', '--tape', action='store_true',
    help="accept tape storage elements when choosing the closest one")
replicate.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
replicate.add_argument('-x', '--bringonline', action='store_true',
    help="do not wait for tape replicas to come online (EXPERT OPTION)")
all_commands.append(replicate)

get = Command('get', t2kdm.interactive.get, "Download file from grid.")
get.add_argument('remotepath', type=str,
    help="the remote logical path, e.g. '/nd280/file.txt'")
get.add_argument('localpath', type=str, nargs='?', default='./',
    help="the local path, e.g. `./`")
get.add_argument('-f', '--force', action='store_true',
    help="overwrite local files if necessary")
get.add_argument('-r', '--recursive', nargs='?', metavar="REGEX", default=False, const=True,
    help="recursively get all files and subdirectories [that match REGEX] of a directory")
get.add_argument('-l', '--list', metavar='FILENAME',
    help="save a list of failed files to FILENAME")
get.add_argument('-s', '--source', type=str, default=None,
    help="the source storage element by name, e.g. 'UKI-SOUTHGRID-RALPP-disk', or by host, e.g. 't2ksrm.nd280.org'. If no source is provided, the replica closest to the destination is chosen")
get.add_argument('-t', '--tape', action='store_true',
    help="accept tape storage elements when choosing the closest one")
get.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
get.add_argument('-x', '--bringonline', action='store_true',
    help="do not wait for tape replicas to come online (EXPERT OPTION)")
all_commands.append(get)

put = Command('put', t2kdm.interactive.put, "Upload file to the grid.")
put.add_argument('localpath', type=str,
    help="the file to be uploaded")
put.add_argument('remotepath', type=str,
    help="the remote logical path, e.g. '/nd280/file.txt'")
put.add_argument('-d', '--destination', type=str, default=None,
    help="the destination storage element by name, e.g. 'UKI-SOUTHGRID-RALPP-disk', or by host, e.g. 't2ksrm.nd280.org'. If no destination is provided, the closest one is chosen")
put.add_argument('-t', '--tape', action='store_true',
    help="accept tape storage elements when choosing the closest one")
put.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
all_commands.append(put)

SEs = Command('SEs', t2kdm.interactive.print_storage_elements, "Print all available storage elements on screen.")
all_commands.append(SEs)

remove = Command('remove', t2kdm.interactive.remove, "Remove file replica from a storage element, if it is not the last one.")
remove.add_argument('remotepath', type=str,
    help="the remote logical path, e.g. '/nd280/file.txt'")
remove.add_argument('destination', type=str,
    help="the destination storage element by name, e.g. 'UKI-SOUTHGRID-RALPP-disk', or by host, e.g. 't2ksrm.nd280.org'")
remove.add_argument('-f', '--final', action='store_true',
    help="do not refuse to remove the last replica of the file, USE WITH CARE!")
remove.add_argument('-r', '--recursive', nargs='?', metavar="REGEX", default=False, const=True,
    help="recursively remove all files and subdirectories [that match REGEX] of a directory")
remove.add_argument('-l', '--list', metavar='FILENAME',
    help="save a list of failed files to FILENAME")
remove.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
remove.add_argument('-x', '--deregister', action='store_true',
    help="deregister only, do NOT try to delete the actual replica (EXPERT OPTION)")
all_commands.append(remove)

fix = Command('fix', t2kdm.interactive.fix, "Try to fix some common issues with a file.")
fix.add_argument('remotepath', type=str,
    help="the remote logical path, e.g. '/nd280/file.txt'")
fix.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
fix.add_argument('-r', '--recursive', nargs='?', metavar="REGEX", default=False, const=True,
    help="recursively fix all files and subdirectories [that match REGEX] of a directory")
fix.add_argument('-l', '--list', metavar='FILENAME',
    help="save a list of failed files to FILENAME")
all_commands.append(fix)

move = Command('move', t2kdm.interactive.move, "Move a file to a new position.",
    epilog="A recursive move only makes sense if the newremotepath is a directory (signified by a '/' at the end).")
move.add_argument('oldremotepath', type=str,
    help="the old remote logical path, e.g. '/nd280/file.txt'")
move.add_argument('newremotepath', type=str,
    help="the new remote logical path, e.g. '/nd280/new_file.txt' or '/new_folder/'")
move.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
move.add_argument('-r', '--recursive', nargs='?', metavar="REGEX", default=False, const=True,
    help="recursively move all files and subdirectories [that match REGEX] of a directory")
move.add_argument('-l', '--list', metavar='FILENAME',
    help="save a list of failed files to FILENAME")
all_commands.append(move)

rename = Command('rename', t2kdm.interactive.rename, "Rename a file using regular expressions",
    epilog="The regular expression is applied to the full path of the file!")
rename.add_argument('remotepath', type=str,
    help="the old remote logical path, e.g. '/nd280/file.txt'")
rename.add_argument('regex_from', type=str,
    help="the regular expression to be replaced, e.g. 't(.)t'")
rename.add_argument('regex_to', type=str,
    help="the regular expression to be put inplace, e.g. 'T\\1T'")
rename.add_argument('-v', '--verbose', action='store_true',
    help="print status messages to the screen")
rename.add_argument('-r', '--recursive', nargs='?', metavar="REGEX", default=False, const=True,
    help="recursively rename all files and subdirectories [that match REGEX] of a directory")
rename.add_argument('-l', '--list', metavar='FILENAME',
    help="save a list of failed files to FILENAME")
all_commands.append(rename)



from setuptools import setup

# Todo: Parse this from a proper readme file in the future
description='T2K Data Manager'
long_description = """t2kdm - T2K Data Manager

Provides functions for a smoother handling of grid data within the T2K
experiment.

"""

setup(name='t2kdm',
    version='1.5.0',
    description=description,
    long_description=long_description,
    url='https://github.com/t2k-software/t2kdm',
    author='Lukas Koch',
    author_email='lukas.koch@mailbox.org',
    license='MIT',
    packages=['t2kdm'],
    install_requires=[
        'sh>=1.12.14',
        'six>=1.10.0',
        'appdirs>=1.4.3',
    ],
    extras_require = {
    },
    python_requires='>=2.6',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    entry_points = {
        'console_scripts': [
            't2kdm-ls=t2kdm.commands:ls.run_from_console',
            't2kdm-replicas=t2kdm.commands:replicas.run_from_console',
            't2kdm-replicate=t2kdm.commands:replicate.run_from_console',
            't2kdm-remove=t2kdm.commands:remove.run_from_console',
            't2kdm-SEs=t2kdm.commands:SEs.run_from_console',
            't2kdm-get=t2kdm.commands:get.run_from_console',
            't2kdm-put=t2kdm.commands:put.run_from_console',
            't2kdm-check=t2kdm.commands:check.run_from_console',
            't2kdm-fix=t2kdm.commands:fix.run_from_console',
            't2kdm-move=t2kdm.commands:move.run_from_console',
            't2kdm-rename=t2kdm.commands:rename.run_from_console',
            't2kdm-cli=t2kdm.cli:run_cli',
            't2kdm-tests=t2kdm.tests:run_tests',
            't2kdm-config=t2kdm.configuration:run_configuration_wizard',
            't2kdm-maid=t2kdm.maid:run_maid',
        ],
    },
    zip_safe=True)



"""Module to deal with regular replication, checking and general housekeeping tasks."""

import argparse
from six.moves import configparser, html_entities
import base64
from six import print_
import t2kdm
import t2kdm.cli
import t2kdm.commands
from contextlib import contextmanager
import sys, os
import sh
from datetime import datetime, timedelta, tzinfo
import posixpath
import tempfile

class UTC(tzinfo):
    """UTC class, because pytz would be overkill"""

    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)
utc = UTC()

def pid_running(pid):
    """Return `True` is a process with the given PID is running."""
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True

class Task(object):
    """Class to organise what need to be done when."""

    def __init__(self, **kwargs):
        """Initialise a basic task.

        All tasks support these basic keyword arguments:

            frequency = 'daily' | 'weekly' | 'monthly'
                How often is the task to be done? Default: 'weekly'

            logfile
                If provided, redirect all output of the task to this file.

        """

        # Handle basic keyword arguments
        self.frequency = kwargs.pop('frequency', 'weekly')
        if self.frequency not in ['daily', 'weekly', 'monthly']:
            raise ValueError("Illegal frequency!")
        self.logfile = kwargs.pop('logfile', None)

        self.last_done = None
        self.state = None

    def get_period(self):
        """Get the time period (1/frequency) of the task."""
        if self.frequency  == 'daily':
            return timedelta(1)
        elif self.frequency  == 'weekly':
            return timedelta(7)
        elif self.frequency  == 'monthly':
            return timedelta(30)
        else:
            raise ValueError("Illegal frequency!")

    @contextmanager
    def redirected_output(self, append=False):
        if self.logfile is not None:
            # Redirect output to logfile
            stdout = sys.stdout
            stderr = sys.stderr
            # Open the file in line buffered mode, so we can read along
            if append:
                mode = 'at'
            else:
                mode= 'wt'
            with open(self.logfile, mode, 1) as f:
                sys.stdout = f
                sys.stderr = f
                try:
                    yield
                finally:
                   sys.stdout = stdout
                   sys.stderr = stdout
        else:
            # Nothing to do
            yield

    def _pre_do(self, id=None):
        """Bookkeeping when starting a task."""
        now = datetime.now(utc)
        self.state = 'STARTED'
        self.last_done = now
        if id is None:
            self.last_id = os.getpid()
        else:
            self.last_id = id

    def _do(self):
        """Internal function that does what needs to be done.

        Must be implemented in inheriting classes.
        Retrns `True` on success or `False` when it failed.
        """
        raise NotImplementedError()

    def do(self, id=None):
        """Actually do the task."""

        with self.redirected_output(append=False):
            self._pre_do(id=id)
            print_(self)
            print_("TASK STARTED")
            # Add a timestamp to the beginning of the output
            sh.date(_out=sys.stdout)

            try:
                success = self._do()
            except Exception as e:
                # Something went wrong
                self._post_do(state='FAILED', id=id)
                print_("TASK FAILED")
                # Add a timestamp to the end of the output
                sh.date(_out=sys.stdout)
                print_(e)
                raise

            if success:
                self._post_do(state='DONE', id=id)
                print_("TASK DONE")
            else:
                self._post_do(state='FAILED', id=id)
                print_("TASK FAILED")
            # Add a timestamp to the end of the output
            sh.date(_out=sys.stdout)

        return success

    def _post_do(self, state='DONE', id=None):
        """Bookkeeping when finishing a task."""
        self.state = state
        if id is None:
            self.last_id = os.getpid()
        else:
            self.last_id = id

    def get_due(self):
        """Calculate how due the task is.

        Return value is a float:

            due < 0     -> Task is not due to be done
            due == 0    -> Task should be done excactly now (on the second!)
            due > 0     -> Task is overdue

        The dueness is scaled by the intended period of task execution,
        i.e. a weekly task that has been run 9 days ago is less due than a daily task
        that has last been run 1.5 days ago.
        """

        now = datetime.now(utc)
        day = 24*3600 # seconds per day
        week = 7*day # seconds per week
        month = 30*day # seconds per month
        T = self.get_period()

        # Everything sucks on python 2.6...
        # Define a function that returns the seconds in a timedelta
        sec = lambda td: float(td.seconds + (td.days * day))

        if self.last_done is None:
            # If no last execution is known, assume the task needs to be done since a month ago
            return month / sec(T)
        else:
            return sec(now - (self.last_done + T)) / sec(T)

    def get_logname(self):
        """Get a valid filename that can be used to log the output of the task."""
        return base64.b64encode(self.get_id(), altchars='+_')+'.txt'

    def get_id(self):
        """Return a string that identifies the task."""
        return str(self)

    def __str__(self):
        """Return a string to identify the task by."""
        return '%s_Task'%(self.frequency,)

class CommandTask(Task):
    """General task based on the commands in t2kdm.commands."""

    def __init__(self, **kwargs):
        self.commandline = kwargs.pop('commandline')
        command, argstr = self.commandline.split(' ', 1)
        for cmd in t2kdm.commands.all_commands:
            if cmd.name == command:
                self.command = cmd
                self.argstr = argstr
                break
        else:
            raise ValueError("Unknown command: %s"%(command,))
        super(CommandTask, self).__init__(**kwargs)

    def _do(self):
        return self.command.run_from_cli(self.argstr, _return=True) == 0

    def __str__(self):
        """Return a string to identify the task by."""
        return '%s_CommandTask_>%s<'%(self.frequency, self.commandline)

class TrimLogTask(Task):
    """Task to trim a log file to a certain number of line.

    The lines will be cut off from the *top*. Equivalent to:

        $ tail -n nlines <path >tempfile
        $ mv tempfile logfile

    """

    def __init__(self, **kwargs):
        """Initialise TrimLogTask.

        Specify the file to be trimmed with the `path` and the number of lines
        to be kept with the `nlines` keyword arguments.
        """

        self.path = kwargs.pop('path')
        self.nlines = kwargs.pop('nlines')
        super(TrimLogTask, self).__init__(**kwargs)

    def _do(self):
        with tempfile.TemporaryFile('w+t') as tf:
            # Write tail of logfile into temporary file
            sh.tail(self.path, lines=self.nlines, _in=self.path, _out=tf)
            # Rewind temporary file
            tf.seek(0)
            # Overwrite old file
            sh.cat(_in=tf, _out=self.path)
        return True

    def __str__(self):
        """Return a string to identify the task by."""
        return '%s_TrimLogTask_of_>%s<_to_>%s<'%(self.frequency, self.path, self.nlines)

class TaskLog(object):
    """Class to handle the logging of task activity."""

    def __init__(self, filename):
        """Use the given filename as log file."""
        self.filename = filename
        self.timeformat = "%Y-%m-%d_%H:%M:%S%z"
        self.id = os.getpid() # Store an ID to identifiy different processes

    def timestamp(self):
        """Return the current timestamp."""
        time = datetime.now(utc)
        return time.strftime(self.timeformat)

    def log(self, state, task, id=None, end='\n'):
        """Log the STARTED, DONE or FAILED of a task.

        Prepends a timestamp and PID.
        """

        if state not in ['STARTED', 'DONE', 'FAILED']:
            return ValueError("Not a valid task state: %s"%(state,))

        if id is None:
            id = self.id
        with open(self.filename, 'at') as f:
            f.write("%s %s %s %s%s"%(self.timestamp(), id, state, task, end))

    class ParseError(Exception):
        pass

    def parse_time(self, timestamp):
        """Return a datetime object according to the timestamp."""
        timeformat = self.timeformat[:-2] # Need to remove '%z' because pthon <3.2 does not understand it
        timestamp = timestamp[:-5] # Same for timestamp, remove '+0000'
        # We just have to assume here that everything is in UTC
        try:
            dt = datetime.strptime(timestamp, timeformat)
        except ValueError:
            raise TaskLog.ParseError()

        # Make timezone-aware
        dt = dt.replace(tzinfo = utc)
        return dt

    def _parse_line(self, line):
        """Return dict of parsed line."""
        elements = line.split()
        if len(elements) < 4:
            raise TaskLog.ParseError()

        ret = {
            'time': self.parse_time(elements[0]),
            'id': elements[1],
            'state': elements[2],
            'task': ' '.join(elements[3:]),
        }
        return ret

    def parse_log(self):
        """Parse the log file and find the last STARTED, DONE and FAILED times of tasks."""

        last_started = {}
        last_done = {}
        last_failed = {}

        with open(self.filename, 'rt') as f:
            for line in f:
                line = line.strip()
                if line.startswith('#'):
                    # Ignore comments
                    continue

                try: # to parse the line for task run information
                    ret = self._parse_line(line)
                except TaskLog.ParseError:
                    continue

                task = ret['task']
                state = ret['state']
                time = ret['time']
                id = ret['id']

                if state == 'STARTED':
                    # Started a task
                    if task not in last_started or time > last_started[task][1]: # Entry is a tuple of (id, time)
                        last_started[task] = (id, time)
                elif state == 'DONE':
                    # Started a task
                    if task not in last_done or time > last_done[task][1]: # Entry is a tuple of (id, time)
                        last_done[task] = (id, time)
                elif state == 'FAILED':
                    # Started a task
                    if task not in last_failed or time > last_failed[task][1]: # Entry is a tuple of (id, time)
                        last_failed[task] = (id, time)

        return last_started, last_done, last_failed

class Maid(object):
    """Class that deals with all regular data keeping tasks.

    It takes care of replication, checksum checks, and regular reports of the state of things.
    """

    def __init__(self, configfile, report=None):
        """Initialise the Maid with the given configuration file.

        The file must look like this:

            [log]
            tasklog = /path/to/logfile.txt
            trimlog = monthly

            [daily]
            replicate /some/folder/ SOME_SE_disk -vr
            replicate /second/different/folder/to/be/replicated/to/the/SE/@ -vr

            [weekly]
            check /some/folder/ -c -s SOME_SE_disk -vr

            [monthly]
            fix /some/other/folder/@ -vr

        Each command in `daily/weekly/monthly` is handled independently. So it is possible to
        request a weekly transfer of a certain folder, while additionally replicating
        a subfolder every day:

            [daily]
            replicate /some/data/folder/very/important/@ SOME_SE_disk -vr

            [monthly]
            replicate /some/data/folder/ SOME_SE_disk -vr

        If the last level of the path to be replicated is an '@' character,
        it will be replaced by the *lexigraphically* last element in the directory.
        This can be used to define a daily replica of the newest data when the
        folder structure is updated with the run numbers.

        If a `report` folder is specified, the output of the tasks will be redirected
        and a browsable webpage generated in that folder.

        """

        self.report = report

        parser = configparser.SafeConfigParser(allow_no_value=True)
        parser.optionxform = str # Need to make options case sensitive
        parser.read(configfile)

        # Store tasks
        self.tasks = {}

        tasklog_path = parser.get('log', 'tasklog')
        self.tasklog = TaskLog(tasklog_path)

        # Create TrimLogTask
        trim_freq = parser.get('log', 'trimlog')
        new_task = TrimLogTask(path=tasklog_path, nlines=1000, frequency=trim_freq)
        new_id = new_task.get_id()
        # If an html report is requested, redirect the output to the appropriate file
        if self.report is not None:
            new_task.logfile = os.path.join(self.report, new_task.get_logname())

        self.tasks[new_id] = new_task

        for sec in parser.sections():
            freq = sec.lower()
            if freq not in ['daily', 'weekly', 'monthly']:
                continue
            print_("Adding %s tasks..."%(freq,))
            for opt in parser.options(sec):
                val = parser.get(sec, opt) # Create the task from the config file line
                if val is None:
                    print_("Adding task: %s"%(opt,))
                else:
                    print_("Adding task: %s = %s"%(opt,val))
                new_task = CommandTask(commandline=opt, frequency=freq)
                new_id = new_task.get_id()
                if new_id in self.tasks: # Make sure the task does not already exist
                    raise RuntimeError("Duplicate task: %s"%(new_id,))
                # If an html report is requested, redirect the output to the appropriate file
                if self.report is not None:
                    new_task.logfile = os.path.join(self.report, new_task.get_logname())

                # Store task in dict of tasks
                self.tasks[new_id] = new_task

    @staticmethod
    def _quote_html(s):
        trans = html_entities.codepoint2name
        ret = ''.join(trans[x] if x in trans else x for x in s)
        return ret

    def generate_index(self):
        """Generate the index.html page for the reports."""

        if self.report is None:
            # Nothing to do
            return
        else:
            indexfile = os.path.join(self.report, 'index.html')

        # Build the html rows for the tasks
        taskrows = ''
        for name in sorted(self.tasks):
            t = self.tasks[name]
            # Task might never have been run
            if t.last_done is None:
                lastrun = "NEVER"
                state = ""
            else:
                lastrun = t.last_done.strftime(self.tasklog.timeformat)
                state = t.state
                if state == 'STARTED':
                    # Check if the PID is actually still there
                    if not pid_running(int(t.last_id)):
                        state = 'STARTED - PID NOT FOUND'

            taskrows += """
                <tr>
                    <td>{lastrun}</td>
                    <td><a href="{logfile}">{name}</a></td>
                    <td>{state}</td>
                </tr>
            """.format(lastrun = lastrun,
                        logfile = t.get_logname(),
                        state = state,
                        name = self._quote_html(t.get_id()))

        context = {
            'timestamp': self.tasklog.timestamp(),
            'taskrows': taskrows,
        }
        with open(indexfile, 'wt') as f:
            f.write("""<!DOCTYPE html>
                <html lang="en">
                  <head>
                    <meta charset="utf-8">
                    <title>T2K Data Manager - Maid Report</title>
                  </head>
                  <body>
                    <h1>T2K Data Manager - Maid Report</h1>
                    <h2>{timestamp}</h2>
                    <table>
                      {taskrows}
                    </table>
                  </body>
                </html>
            """.format(**context))

    def do_task(self, task):
        """Do a specific task and log it in the tasklog.

        Return `True` if succesfull.
        """

        # Report if necessary
        self.update_task_states() # Make sure tasks are up to date
        task._pre_do() # Set task to STARTED
        self.generate_index()

        # Do the task and log it
        self.tasklog.log('STARTED', task.get_id())
        try:
            success = task.do()
        except KeyboardInterrupt:
            # Catch keyboard interrupts and exit gracefully
            self.tasklog.log('FAILED', task.get_id())
            success = False
        except:
            # No idea what happened
            # Just fail
            self.tasklog.log('FAILED', task.get_id())
            raise
        else:
            # No exceptions, so go by the return value
            if success:
                self.tasklog.log('DONE', task.get_id())
            else:
                self.tasklog.log('FAILED', task.get_id())

        # Report if necessary
        self.update_task_states() # Update tasks that might have run in parallel
        self.generate_index()

        return success

    def update_task_states(self):
        """Read log and update when tasks were last done."""

        started, done, failed = self.tasklog.parse_log()

        for task in started:
            if task in self.tasks:
                t = self.tasks[task]
                if t.last_done is None or t.last_done < started[task][1]:
                    # Do not overwrite start dates that were updated in this process
                    t.last_done = started[task][1] # tuple of (pid, time)
                    t.last_id =  started[task][0]
                    t.state = 'STARTED'

        for task in done:
            if task in self.tasks:
                t = self.tasks[task]
                if t.last_done is None:
                    t.last_done = done[task][1] # tuple of (pid, time)
                    t.last_id =  done[task][0]
                    t.state = 'DONE'
                elif t.last_done <= done[task][1]:
                    # Do not overwrite if task was started after last DONE
                    t.last_id =  done[task][0]
                    t.state = 'DONE'

        for task in failed:
            if task in self.tasks:
                t = self.tasks[task]
                if t.last_done is None:
                    t.last_done = failed[task][1] # tuple of (pid, time)
                    t.last_id =  failed[task][0]
                    t.state = 'FAILED'
                elif t.last_done <= failed[task][1]:
                    # Do not overwrite if task was started after last FAILED
                    t.last_id =  failed[task][0]
                    t.state = 'FAILED'

    def get_open_tasks(self, return_all=False):
        """Return a list of open tasks in order of how due they are.

        If `return_all` is `True`, all tasks will be returned, not just the due ones.
        """

        ret = []
        for t in self.tasks:
            task = self.tasks[t]
            if return_all or task.get_due() >= 0:
                ret.append(task)

        # Sort tasks by dueness
        ret.sort(key=lambda tsk: tsk.get_due(), reverse=True) # Highest due on top
        return ret

    def do_something(self, eager=False):
        """Find an open task and do it.

        If `eager` is `True`, do tasks even before they are due again.
        """

        self.update_task_states()
        tasks = self.get_open_tasks(return_all=eager)

        if len(tasks) > 0:
            print_("Due tasks:")
            for t in tasks:
                print_("* %s (%.3f)"%(t, t.get_due()))

            for t in tasks:
                if t.state == 'STARTED' and pid_running(int(t.last_id)):
                    print_("%s seems to be running already. Skipping..."%(t,))
                    continue
                else:
                    # Found a task we should do
                    break
            else:
                print_("All due tasks seem to be running already. Nothing to do.")
                return

            print_("Starting %s..."%(t))
            if self.do_task(t):
                print_("Done.")
            else:
                print_("Failed.")
        else:
            print_("Nothing to do.")

def run_maid():
    """Start the Maid program and do some tasks.

    Intended to be run multiple times per day, but at least daily.
    """

    parser = argparse.ArgumentParser(description="Regular housekeeping for the T2K data. Run at least daily!")
    parser.add_argument('-e', '--eager', action='store_true',
                        help="do a task, even if it is not due yet")
    parser.add_argument('-r', '--report', metavar='FOLDER', default=None,
                        help="generate an html report in the given folder")
    args = parser.parse_args()

    maid = Maid(t2kdm.config.maid_config, report=args.report)
    maid.do_something(eager=args.eager)

if __name__ == '__main__':
    run_maid()
